import json
from pathlib import Path
import pytest

from erdb.utils.attack_power import Attributes, CalculatorData, ArmamentCalculator


"""
Version of the game the sample data has been explicitly collected for
"""
_GAME_VERSION: str = "1.05.0"

_TEST_DIR = Path(__file__).parent.resolve()
_DATA_DIR = _TEST_DIR / "attack_values"
_GAMEDATA_DIR = _TEST_DIR.parent / _GAME_VERSION
_ATTRIBUTE_SETS = [Attributes.from_string(f.stem) for f in _DATA_DIR.glob("*")]

def pytest_generate_tests(metafunc):
    assert "armament" in metafunc.fixturenames
    assert "attribs" in metafunc.fixturenames

    with open(_GAMEDATA_DIR / "armaments.json") as f:
        armaments = json.load(f)["Armaments"]

    armament_data: list[tuple[str, str]] = []
    armament_ids: list[str] = []

    for armament, properties in armaments.items():
        for affinity in properties["affinity"].keys():
            armament_data.append((affinity, armament))
            armament_ids.append(f"{affinity} {armament}")

    metafunc.parametrize("armament", armament_data, ids=armament_ids)
    metafunc.parametrize("attribs", _ATTRIBUTE_SETS, ids=[*map(str, _ATTRIBUTE_SETS)])

@pytest.fixture(scope="module")
def calc_data() -> CalculatorData:
    return CalculatorData.create(_GAMEDATA_DIR)

@pytest.fixture(scope="module")
def results_data() -> dict[str, dict]:
    def load(attribs):
        with open(_DATA_DIR / f"{attribs}.json") as f:
            return json.load(f)

    return {attribs: load(attribs) for attribs in _ATTRIBUTE_SETS}

def test_attack_power(calc_data, armament, attribs, results_data):
    affinity, name = armament
    level = "0" if name == "Meteorite Staff" else "10"

    expected = results_data[attribs][name][affinity]["attack_power"]
    calc = ArmamentCalculator(calc_data, name, affinity, level)
    ap = calc.attack_power(attribs)

    if ap.physical.total > 0:
        assert len(expected["physical"]) == 4
        assert expected["physical"][3] == ap.physical.total

    if ap.magic.total > 0:
        assert len(expected["magic"]) == 4
        assert expected["magic"][3] == ap.magic.total

    if ap.fire.total > 0:
        assert len(expected["fire"]) == 4
        assert expected["fire"][3] == ap.fire.total

    if ap.lightning.total > 0:
        assert len(expected["lightning"]) == 4
        assert expected["lightning"][3] == ap.lightning.total

    if ap.holy.total > 0:
        assert len(expected["holy"]) == 4
        assert expected["holy"][3] == ap.holy.total


from typing import Dict, List, NamedTuple, Self, Tuple, Union

from erdb.typing.effects import EffectModel, EffectType, AttributeName, SchemaEffect
from erdb.typing.enums import SpEffectType, AttackType, AttackCondition


class SpEffectConditionOffset(NamedTuple):
    condition: Union[None, SpEffectType, AttackType, AttackCondition]
    offset: int

class SpEffectRanges(NamedTuple):
    class IntRange(NamedTuple):
        begin: int
        end: int

        def __contains__(self, __x: object) -> bool:
            assert isinstance(__x, int)
            return self.begin <= __x <= self.end

    ranges: List[IntRange]

    def __contains__(self, __x: object) -> bool:
        if isinstance(__x, str):
            return int(__x) in self

        if not isinstance(__x, int):
            return False

        return any(__x in r for r in self.ranges)

    @classmethod
    def construct(cls, *ranges: Tuple[int, int]) -> Self:
        return cls([cls.IntRange(r[0], r[1]) for r in ranges])

"""
Some SpEffects don't seem to have anything that identify what they do (Greatshield Talisman)
and/or are simply too difficult to parse (Winged Sword Insignia). They are listed here for visibility.

The list for each SpEffect ID should contain only the problematic effect(s). If any effect is parsable,
it will be appended to this list automatically during generation.

!! THIS LIST MUST BE CONFIRMED AFTER EACH UPDATE !!
Last confirmed ER version: 1.04.1
"""
_FROM_ID: Dict[int, List[SchemaEffect]] = {
    # >> Greatshield Talisman
    # This effect does not seem to utilize `guardStaminaCutRate`, unlike other shield buffs.
    341000: [
        SchemaEffect(
            attribute=AttributeName.STABILITY,
            effect_model=EffectModel.MULTIPLICATIVE,
            effect_type=EffectType.POSITIVE,
            value=1.1
        )
    ],
    # >> Concealing Veil
    # Its only identifying property is the unique `invocationConditionsStateChange1` "Trigger on Crouch"
    360100: [
        SchemaEffect(
            attribute=AttributeName.INVISIBLE_AT_DISTANCE,
            conditions=[str(SpEffectType.TRIGGER_ON_CROUCH)],
            effect_model=EffectModel.ADDITIVE,
            effect_type=EffectType.POSITIVE,
            value=1
        )

    ],
    # >> Furled Finger's Trick-Mirror
    # Nothing is identifying this SpEffect besides its ID
    360800: [
        SchemaEffect(
            attribute=AttributeName.APPEAR_AS_HOST,
            effect_model=EffectModel.ADDITIVE,
            effect_type=EffectType.NEUTRAL,
            value=1
        )
    ],
    # >> Host's Trick-Mirror
    # Nothing is identifying this SpEffect besides its ID
    360900: [
        SchemaEffect(
            attribute=AttributeName.APPEAR_AS_COOPERATOR,
            effect_model=EffectModel.ADDITIVE,
            effect_type=EffectType.NEUTRAL,
            value=1
        )
    ],
    # >> Shabriri's Woe
    # Nothing is identifying this SpEffect besides its ID
    360500: [
        SchemaEffect(
            attribute=AttributeName.ATTRACT_ENEMY_AGGRESSION,
            effect_model=EffectModel.ADDITIVE,
            effect_type=EffectType.NEUTRAL,
            value=1
        )
    ],
}

"""
Some SpEffects reference other SpEffects by themselves or their stateInfo. Some of them can be parsed,
while others can be problematic. This is the list of problematic SpEffects which hardcodes the offsets
of the effects they are referencing, along with a potential condition which activates them.
"""
_FROM_OFFSET: Dict[int, list[SpEffectConditionOffset]] = {
    350400: [ # Godskin Swaddling Cloth
        SpEffectConditionOffset(AttackCondition.SUCCESSIVE_HITS, 1)
    ],
    312500: [ # Millicent's Prosthesis
        SpEffectConditionOffset(AttackCondition.SUCCESSIVE_3_HITS, 5),
        SpEffectConditionOffset(AttackCondition.SUCCESSIVE_6_HITS, 6),
        SpEffectConditionOffset(AttackCondition.SUCCESSIVE_9_HITS, 7),
    ],
    320800: [ # Winged Sword Insignia
        SpEffectConditionOffset(AttackCondition.SUCCESSIVE_3_HITS, 4),
        SpEffectConditionOffset(AttackCondition.SUCCESSIVE_6_HITS, 5),
        SpEffectConditionOffset(AttackCondition.SUCCESSIVE_9_HITS, 6),
    ],
    320810: [ # Rotten Winged Sword Insignia
        SpEffectConditionOffset(AttackCondition.SUCCESSIVE_3_HITS, 4),
        SpEffectConditionOffset(AttackCondition.SUCCESSIVE_6_HITS, 5),
        SpEffectConditionOffset(AttackCondition.SUCCESSIVE_9_HITS, 6),
    ],
    350500: [ # Assassin's Crimson Dagger
        SpEffectConditionOffset(None, 2) # None -- the referenced effect contains conditions
    ],
    350600: [ # Assassin's Cerulean Dagger
        SpEffectConditionOffset(None, 2) # None -- the referenced effect contains conditions
    ]
}

"""
Some SpEffects types define effects themselves without affecting any attributes.
"""
_FROM_TYPE: Dict[SpEffectType, List[SchemaEffect]] = {
    # >> Crucible Knot Talisman + IDs: 9648, 6044000
    # Crucible Knot Talisman does not seem to define exactly how much the damage is reduced.
    SpEffectType.REDUCE_HEADSHOT_IMPACT: [
        SchemaEffect(
            attribute=AttributeName.REDUCE_HEADSHOT_IMPACT,
            effect_model=EffectModel.ADDITIVE,
            effect_type=EffectType.POSITIVE,
            value=1
        )
    ],
    # >> Entwining Umbilical Cord
    SpEffectType.SWITCH_ANIMATION_GENDER: [
        SchemaEffect(
            attribute=AttributeName.SWITCH_ANIMATION_GENDER,
            effect_model=EffectModel.ADDITIVE,
            effect_type=EffectType.NEUTRAL,
            value=1
        )
    ],
    # >> Sacrificial Twig
    SpEffectType.DESTROY_ACCESSORY_BUT_SAVE_RUNES: [
        SchemaEffect(
            attribute=AttributeName.PRESERVE_RUNES_ON_DEATH,
            effect_model=EffectModel.ADDITIVE,
            effect_type=EffectType.POSITIVE,
            value=1
        ),
        SchemaEffect(
            attribute=AttributeName.DESTROY_ITEM_ON_DEATH,
            effect_model=EffectModel.ADDITIVE,
            effect_type=EffectType.NEGATIVE,
            value=1
        ),
    ],
}

def get(index: int, sp_effect_type) -> List[SchemaEffect]:
    return _FROM_ID.get(index, []) + _FROM_TYPE.get(sp_effect_type, [])

def get_conditions(index: int) -> List[SpEffectConditionOffset]:
    return _FROM_OFFSET.get(index, [])

def get_status_effect_ranges() -> SpEffectRanges:
    return SpEffectRanges.construct((6400, 6810), (105000, 109000))


from erdb.typing.models.talisman import Talisman
from erdb.typing.models.effect import Effect
from erdb.typing.models import NonEmptyStr
from erdb.effect_parser import parse_effects
from erdb.typing.params import ParamRow
from erdb.typing.enums import ItemIDFlag
from erdb.typing.api_version import ApiVersion
from erdb.table._retrievers import ParamDictRetriever, MsgsRetriever, RetrieverData
from erdb.table._common import RowPredicate, TableSpecContext


class TalismanTableSpec(TableSpecContext):
    model = {
        ApiVersion.VER_1: Talisman,
    }

    main_param_retriever = ParamDictRetriever("EquipParamAccessory", ItemIDFlag.ACCESSORIES)

    predicates: list[RowPredicate] = [
        lambda row: 1000 <= row.index < 999999,
    ]

    param_retrievers = {
        "effects": ParamDictRetriever("SpEffectParam", ItemIDFlag.NON_EQUIPABBLE, id_min=310000, id_max=400000)
    }

    msg_retrievers = {
        "names": MsgsRetriever("AccessoryName"),
        "summaries": MsgsRetriever("AccessoryInfo"),
        "descriptions": MsgsRetriever("AccessoryCaption")
    }

    @classmethod
    def _find_conflicts(cls, data: RetrieverData, group: int) -> list[NonEmptyStr]:
        def valid(row: ParamRow) -> bool:
            return row["accessoryGroup"].as_int == group and row.index < 9999999
        return [NonEmptyStr(cls.parse_name(data.msgs["names"][row.index])) for row in data.main_param.values() if valid(row)]

    @classmethod
    def make_object(cls, api: ApiVersion, data: RetrieverData, row: ParamRow):
        effects = data.params["effects"]

        return Talisman(
            **cls.make_item(data, row),
            **cls.make_contrib(data, row, "locations", "remarks"),
            weight=row["weight"].as_float,
            effects=[Effect(**eff) for eff in parse_effects(row, effects, "refId")],
            conflicts=cls._find_conflicts(data, row["accessoryGroup"].as_int),
        )


from typing import NamedTuple

from erdb.typing.game_version import GameVersion
from erdb.loaders.params import load as load_params, load_ids as load_param_ids, load_msg
from erdb.loaders.contrib import load as load_contrib
from erdb.typing.params import ParamDict
from erdb.typing.enums import ItemIDFlag
from erdb.shop import Lookup


class RetrieverData(NamedTuple):
    main_param: ParamDict
    params: dict[str, ParamDict]
    msgs: dict[str, dict[int, str]]
    shops: dict[str, Lookup]
    contrib: dict[str, dict]

class ParamDictRetriever(NamedTuple):
    param_name: str
    item_id_flag: ItemIDFlag
    id_min: int | None = None
    id_max: int | None = None

    def get(self, version: GameVersion) -> ParamDict:
        args = [self.param_name, version, self.item_id_flag]
        args += [arg for arg in [self.id_min, self.id_max] if arg is not None]
        func = load_params if len(args) <= 3 else load_param_ids
        return func(*args) # type: ignore

    def __contains__(self, __x: object) -> bool:
        assert isinstance(__x, int), f"Can only check for integer range"
        return  (not self.id_min or self.id_min <= __x) \
            and (not self.id_max or self.id_max >= __x)

class MsgsRetriever(NamedTuple):
    file_name: str

    def get(self, version: GameVersion) -> dict[int, str]:
        return load_msg(self.file_name, version)

class ShopRetriever(NamedTuple):
    shop_lineup_id_min: int | None
    shop_lineup_id_max: int | None
    material_set_id_min: int | None
    material_set_id_max: int | None
    recipe: bool = False

    def get(self, version: GameVersion) -> Lookup:
        F = ParamDictRetriever
        shop_param = "ShopLineupParam_Recipe" if self.recipe else "ShopLineupParam"
        shop = F(shop_param, ItemIDFlag.NON_EQUIPABBLE, self.shop_lineup_id_min, self.shop_lineup_id_max)
        mats = F("EquipMtrlSetParam", ItemIDFlag.NON_EQUIPABBLE, self.material_set_id_min, self.material_set_id_max)
        return Lookup(shop.get(version), mats.get(version))

class ContribRetriever(NamedTuple):
    def get(self, element_name: str, version: GameVersion) -> dict[str, dict]:
        return load_contrib(element_name, version)


from pydantic import Field
from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config, NonEmptyStr
from erdb.typing.models.item import Item
from erdb.typing.categories import CraftingMaterialCategory


@dataclass(config=dt_config())
class CraftingMaterial(Item):
    category: CraftingMaterialCategory = Field(...,
        description="Crafting Material category to discern its use.",
        example=CraftingMaterialCategory.FAUNA,
    )
    hint: NonEmptyStr = Field(...,
        description="In-game hint on where to find the Crafting Material.",
        example="Found by hunting particularly large beasts",
    )
    products: list[NonEmptyStr] = Field(...,
        description="List of crafting products of the Crafting Material.",
        min_items=1,
        example=[
            "Beastlure Pot",
            "Exalted Flesh",
            "Bone Great Arrow (Fletched)",
            "Bone Great Arrow",
            "Bone Ballista Bolt",
        ]
    )


from enum import Enum
from typing import Self

from erdb.typing.enums import GoodsSortGroupID, GoodsType
from erdb.typing.params import ParamRow


class _CategoryBase(str, Enum):
    @classmethod
    def get(cls, row: ParamRow) -> Self | None:
        assert False, "not implemented"

    @classmethod
    def from_row(cls, row: ParamRow) -> Self:
        if cat := cls.get(row):
            return cat
        raise KeyError

class AmmoCategory(_CategoryBase):
    ARROW = "Arrow"
    GREATARROW = "Greatarrow"
    BOLT = "Bolt"
    GREATBOLT = "Greatbolt"

    @classmethod
    def get(cls, row: ParamRow) -> Self | None:
        return {
            81: AmmoCategory.ARROW,
            83: AmmoCategory.GREATARROW,
            85: AmmoCategory.BOLT,
            86: AmmoCategory.GREATBOLT,
        }.get(row["wepType"].as_int)

class ArmamentCategory(_CategoryBase):
    DAGGER = "Dagger"
    STRAIGHT_SWORD = "Straight Sword"
    GREATSWORD = "Greatsword"
    COLOSSAL_SWORD = "Colossal Sword"
    CURVED_SWORD = "Curved Sword"
    CURVED_GREATSWORD = "Curved Greatsword"
    KATANA = "Katana"
    TWINBLADE = "Twinblade"
    THRUSTING_SWORD = "Thrusting Sword"
    HEAVY_THRUSTING_SWORD = "Heavy Thrusting Sword"
    AXE = "Axe"
    GREATAXE = "Greataxe"
    HAMMER = "Hammer"
    GREAT_HAMMER = "Great Hammer"
    FLAIL = "Flail"
    SPEAR = "Spear"
    GREAT_SPEAR = "Great Spear"
    HALBERD = "Halberd"
    REAPER = "Reaper"
    FIST = "Fist"
    CLAW = "Claw"
    WHIP = "Whip"
    COLOSSAL_WEAPON = "Colossal Weapon"
    LIGHT_BOW = "Light Bow"
    BOW = "Bow"
    GREATBOW = "Greatbow"
    CROSSBOW = "Crossbow"
    BALLISTA = "Ballista"
    GLINTSTONE_STAFF = "Glintstone Staff"
    SACRED_SEAL = "Sacred Seal"
    SMALL_SHIELD = "Small Shield"
    MEDIUM_SHIELD = "Medium Shield"
    GREATSHIELD = "Greatshield"
    TORCH = "Torch"

    @classmethod
    def get(cls, row: ParamRow) -> Self | None:
        return _ARMAMENT_CATEGORY_IDS.get(row["wepType"].as_int)

    @property
    def ingame(self) -> str:
        return _ARMAMENT_CATEGORY_INGAME[self]

class ArmorCategory(_CategoryBase):
    HEAD = "Head"
    BODY = "Body"
    ARMS = "Arms"
    LEGS = "Legs"

    @classmethod
    def get(cls, row: ParamRow) -> Self | None:
        return {
            0: ArmorCategory.HEAD,
            1: ArmorCategory.BODY,
            2: ArmorCategory.ARMS,
            3: ArmorCategory.LEGS,
        }.get(row["protectorCategory"].as_int)

class BolsteringMaterialCategory(_CategoryBase):
    FLASK = "Flask"
    SMITHING_STONE = "Smithing Stone"
    SOMBER_SMITHING_STONE = "Somber Smithing Stone"
    GLOVEWORT = "Glovewort"

    @classmethod
    def get(cls, row: ParamRow) -> Self | None:
        G = GoodsSortGroupID
        B = BolsteringMaterialCategory
        return {
            G.GROUP_1: B.FLASK,
            G.GROUP_2: B.SMITHING_STONE,
            G.GROUP_3: B.SOMBER_SMITHING_STONE,
            G.GROUP_4: B.GLOVEWORT,
        }.get(G(row["sortGroupId"].as_int))

class CraftingMaterialCategory(_CategoryBase):
    FAUNA = "Fauna"
    FLORA = "Flora"
    OBJECT = "Object"

    @classmethod
    def get(cls, row: ParamRow) -> Self | None:
        G = GoodsSortGroupID
        B = CraftingMaterialCategory
        return {
            G.GROUP_1: B.FAUNA,
            G.GROUP_2: B.FLORA,
            G.GROUP_3: B.OBJECT,
        }.get(G(row["sortGroupId"].as_int))

class InfoCategory(_CategoryBase):
    PAINTING = "Painting"
    NOTE = "Note"
    TUTORIAL = "Tutorial"
    CLUE = "Clue"

    @classmethod
    def get(cls, row: ParamRow) -> Self | None:
        G = GoodsSortGroupID
        I = InfoCategory
        return {
            G.GROUP_1: I.PAINTING if "Painting" in row.name else I.NOTE if "Note" in row.name else I.CLUE,
            G.GROUP_2: I.TUTORIAL,
        }.get(G(row["sortGroupId"].as_int))

class KeyCategory(_CategoryBase):
    GREAT_RUNE = "Great Rune"
    MENDING_RUNE = "Mending Rune"
    CONTAINER = "Container"
    EXPLORATION = "Exploration"
    QUEST = "Quest"
    EXCHANGE = "Exchange"
    FEATURE = "Feature"
    WHETBLADE = "Whetblade"
    MAP = "Map"

    @classmethod
    def get(cls, row: ParamRow) -> Self | None:
        K = KeyCategory
        G = GoodsSortGroupID

        for custom in (K.GREAT_RUNE, K.MENDING_RUNE, K.WHETBLADE):
            if custom.value in row.name:
                return custom

        return {
            G.GROUP_3: K.CONTAINER,
            G.GROUP_4: K.EXPLORATION,
            G.GROUP_5: K.EXCHANGE if row["isConsume"].as_bool else K.QUEST,
            G.GROUP_6: K.FEATURE,
            G.GROUP_7: K.MAP,
        }.get(G(row["sortGroupId"].as_int))

class ShopCategory(_CategoryBase):
    COOKBOOK = "Cookbook"
    BELL_BEARING = "Bell Bearing"
    SPELLBOOK = "Spellbook"

    @classmethod
    def get(cls, row: ParamRow) -> Self | None:
        G = GoodsSortGroupID
        S = ShopCategory
        return {
            G.GROUP_6: S.COOKBOOK,
            G.GROUP_8: S.BELL_BEARING,
            G.GROUP_9: S.BELL_BEARING,
            G.GROUP_10: S.SPELLBOOK,
        }.get(G(row["sortGroupId"].as_int))

class SpellCategory(_CategoryBase):
    INCANTATION = "Incantation"
    SORCERY = "Sorcery"

    @classmethod
    def get(cls, row: ParamRow) -> Self | None:
        return {
            0: SpellCategory.SORCERY,
            1: SpellCategory.INCANTATION,
        }.get(row["ezStateBehaviorType"].as_int)

class ToolCategory(_CategoryBase):
    ESSENTIAL = "Essential"
    EDIBLE = "Edible"
    POT = "Pot"
    AROMATIC = "Aromatic"
    THROWABLE = "Throwable"
    OFFENSIVE = "Offensive"
    GREASE = "Grease"
    UTILITY = "Utility"
    GOLDEN_RUNE = "Golden Rune"
    GREAT_RUNE = "Great Rune"
    REMEMBERANCE = "Rememberance"
    CRYSTAL_TEAR = "Crystal Tear"
    ONLINE = "Online"

    @classmethod
    def get(cls, row: ParamRow) -> Self | None:
        G = GoodsSortGroupID
        T = ToolCategory

        goods_type = row["goodsType"]

        if goods_type == GoodsType.WONDROUS_PHYSICK_TEAR:
            return T.CRYSTAL_TEAR

        if goods_type == GoodsType.GREAT_RUNE:
            return T.GREAT_RUNE

        return {
            G.GROUP_1: T.ESSENTIAL,
            G.GROUP_2: T.EDIBLE,
            G.GROUP_3: T.POT,
            G.GROUP_4: T.AROMATIC,
            G.GROUP_5: T.THROWABLE,
            G.GROUP_6: T.OFFENSIVE,
            G.GROUP_7: T.GREASE,
            G.GROUP_8: T.UTILITY,
            G.GROUP_9: T.UTILITY,
            G.GROUP_10: T.GOLDEN_RUNE,
            G.REMEMBERANCES: T.REMEMBERANCE,
            G.ONLINE: T.ONLINE
        }.get(G(row["sortGroupId"].as_int))

_ARMAMENT_CATEGORY_IDS: dict[int, ArmamentCategory] = {
    1: ArmamentCategory.DAGGER,
    3: ArmamentCategory.STRAIGHT_SWORD,
    5: ArmamentCategory.GREATSWORD,
    7: ArmamentCategory.COLOSSAL_SWORD,
    9: ArmamentCategory.CURVED_SWORD,
    11: ArmamentCategory.CURVED_GREATSWORD,
    13: ArmamentCategory.KATANA,
    14: ArmamentCategory.TWINBLADE,
    15: ArmamentCategory.THRUSTING_SWORD,
    16: ArmamentCategory.HEAVY_THRUSTING_SWORD,
    17: ArmamentCategory.AXE,
    19: ArmamentCategory.GREATAXE,
    21: ArmamentCategory.HAMMER,
    23: ArmamentCategory.GREAT_HAMMER,
    24: ArmamentCategory.FLAIL,
    25: ArmamentCategory.SPEAR,
    28: ArmamentCategory.GREAT_SPEAR,
    29: ArmamentCategory.HALBERD,
    31: ArmamentCategory.REAPER,
    35: ArmamentCategory.FIST,
    37: ArmamentCategory.CLAW,
    39: ArmamentCategory.WHIP,
    41: ArmamentCategory.COLOSSAL_WEAPON,
    50: ArmamentCategory.LIGHT_BOW,
    51: ArmamentCategory.BOW,
    53: ArmamentCategory.GREATBOW,
    55: ArmamentCategory.CROSSBOW,
    56: ArmamentCategory.BALLISTA,
    57: ArmamentCategory.GLINTSTONE_STAFF,
    61: ArmamentCategory.SACRED_SEAL,
    65: ArmamentCategory.SMALL_SHIELD,
    67: ArmamentCategory.MEDIUM_SHIELD,
    69: ArmamentCategory.GREATSHIELD,
    87: ArmamentCategory.TORCH,
}

_ARMAMENT_CATEGORY_INGAME: dict[ArmamentCategory, str] = {
    ArmamentCategory.DAGGER: "Dagger",
    ArmamentCategory.STRAIGHT_SWORD: "SwordNormal",
    ArmamentCategory.GREATSWORD: "SwordLarge",
    ArmamentCategory.COLOSSAL_SWORD: "SwordGigantic",
    ArmamentCategory.CURVED_SWORD: "SaberNormal",
    ArmamentCategory.CURVED_GREATSWORD: "SaberLarge",
    ArmamentCategory.KATANA: "katana", # yes, lowercase
    ArmamentCategory.TWINBLADE: "SwordDoubleEdge",
    ArmamentCategory.THRUSTING_SWORD: "SwordPierce",
    ArmamentCategory.HEAVY_THRUSTING_SWORD: "RapierHeavy",
    ArmamentCategory.AXE: "AxeNormal",
    ArmamentCategory.GREATAXE: "AxeLarge",
    ArmamentCategory.HAMMER: "HammerNormal",
    ArmamentCategory.GREAT_HAMMER: "HammerLarge",
    ArmamentCategory.FLAIL: "Flail",
    ArmamentCategory.SPEAR: "SpearNormal",
    # unused spear: "SpearLarge"
    ArmamentCategory.GREAT_SPEAR: "SpearHeavy",
    ArmamentCategory.HALBERD: "SpearAxe",
    ArmamentCategory.REAPER: "Sickle",
    ArmamentCategory.FIST: "Knuckle",
    ArmamentCategory.CLAW: "Claw",
    ArmamentCategory.WHIP: "Whip",
    ArmamentCategory.COLOSSAL_WEAPON: "AxhammerLarge",
    ArmamentCategory.LIGHT_BOW: "BowSmall",
    ArmamentCategory.BOW: "BowNormal",
    ArmamentCategory.GREATBOW: "BowLarge",
    ArmamentCategory.CROSSBOW: "ClossBow", # clossbow...
    ArmamentCategory.BALLISTA: "Ballista",
    ArmamentCategory.GLINTSTONE_STAFF: "Staff",
    # unused catalyst: "Sorcery"
    ArmamentCategory.SACRED_SEAL: "Talisman",
    ArmamentCategory.SMALL_SHIELD: "ShieldSmall",
    ArmamentCategory.MEDIUM_SHIELD: "ShieldNormal",
    ArmamentCategory.GREATSHIELD: "ShieldLarge",
    ArmamentCategory.TORCH: "Torch",
}


from pydantic import Field, NonNegativeInt, PositiveInt
from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config
from erdb.typing.models.item import Item
from erdb.typing.models.common import StatRequirements
from erdb.typing.categories import SpellCategory
from erdb.typing.enums import SpellHoldAction


@dataclass(config=dt_config())
class Spell(Item):
    fp_cost: NonNegativeInt = Field(...,
        description="Cost of FP to cast.",
        example=40,
    )
    fp_cost_extra: NonNegativeInt = Field(...,
        description="Additional cost of FP for a charged/continued attack. This is on top of the regular cost, not the full cost. Always 0 for non-holdable Spells.",
        example=10,
    )
    sp_cost: NonNegativeInt = Field(...,
        description="Cost of stamina to cast.",
        example=34,
    )
    sp_cost_extra: NonNegativeInt = Field(...,
        description="Additional cost of stamina for a charged/continued attack. This is on top of the regular cost, not the full cost. Always 0 for non-holdable Spells.",
        example=0
    )
    category: SpellCategory = Field(...,
        description="Category of the Spell.",
        example=SpellCategory.SORCERY,
    )
    slots_used: PositiveInt = Field(...,
        description="Specifies the number of Memory Slots occupied.",
        example=3,
    )
    hold_action: SpellHoldAction = Field(...,
        description="Defines the behavior of the Spell when the cast is held.",
        example=SpellHoldAction.CONTINUOUS,
    )
    is_weapon_buff: bool = Field(...,
        description="Specifies whether the Spell buffs a weapon in the right hand.",
        example=False,
    )
    is_shield_buff: bool = Field(...,
        description="Specifies whether the Spell buffs a shield in the left hand.",
        example=False,
    )
    is_horseback_castable: bool = Field(...,
        description="Specifies whether the Spell can be cast while on horseback.",
        example=True,
    )
    requirements: StatRequirements = Field(...,
        description="Attribute requirements of the Spell.",
        example=StatRequirements(intelligence=60),
    )